(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
header {* Ceta Simplifier *}

theory Ceta_Simplifier
imports Ceta
uses ("../generated/SML/EvalCeta.ML")
begin

export_code certify_proof parse_cert_problem termifyCR termifyCP in Eval module_name EvalCeta file "../generated/SML/EvalCeta.ML"

use "../generated/SML/EvalCeta.ML"


ML {*
  val input = File.read (Path.explode "~/rewriting/examples/rational_matrix.proof.xml") |> String.explode
  val parse = EvalCeta.parse_cert_problem
  val ceta = EvalCeta.certify_proof true
  val res = (parse input, ceta input)
*}

ML {*
 fun compute_result file = File.read (Path.explode file)
   |> String.explode
   |> EvalCeta.certify_proof true
   |> EvalCeta.termifyCR
 fun compute_proof file = File.read (Path.explode file)
   |> String.explode
   |> EvalCeta.parse_cert_problem
   |> EvalCeta.termifyCP
 fun import_result file =
   Local_Theory.define ((@{binding my_result}, NoSyn), ((@{binding my_result_def}, []), compute_result file)) #> snd
 fun import_proof file =
   Local_Theory.define ((@{binding my_proof}, NoSyn), ((@{binding my_proof_def}, []), compute_proof file)) #> snd
*}

local_setup {*
 import_result "~/rewriting/examples/rational_matrix.proof.xml"
*}

thm my_result_def

local_setup {*
 import_proof "~/rewriting/examples/div.proof.xml"
*}

thm my_proof_def
                   
fun sn_via_ceta :: "(string,string)rules \<Rightarrow> (string,label_type,string)trs_proof \<Rightarrow> bool" where 
  "sn_via_ceta r prf = (case (certify_cert_problem False tp_impl dpp_impl (TRS_Proof No_Strategy (init_lab_rules r :: ((string,label_type)lab,string)rules) None prf)) of
               Certified _ \<Rightarrow> True
             | _ \<Rightarrow> False)"

lemma sn_via_ceta: "sn_via_ceta r prf \<Longrightarrow> SN (rstep (set r))"
proof -
  let ?r = "init_lab_rules r :: ((string,label_type)lab,string)rules"
  let ?r' = "init_lab_trs (set r)::((string, label_type) lab, string) trs"
  assume "sn_via_ceta r prf"
  then obtain x where "certify_cert_problem False tp_impl dpp_impl (TRS_Proof No_Strategy ?r None prf) = Certified x" 
    by (cases "certify_cert_problem False tp_impl dpp_impl (TRS_Proof No_Strategy ?r None prf)", auto)
  from certify_cert_problem_sound[OF tp_impl dpp_impl this] have
    "SN (rstep (set ?r))" by simp
  hence SN: "SN (rstep ?r')"
    unfolding map_trs.simps by auto
  {
    fix t
    have "SN_on (rstep (set r)) {t}"
      by (rule SN_map_imp_SN[OF SN[unfolded SN_def, THEN spec]])
  }
  from SN_I[OF this] show ?thesis .
qed


fun extract_proof where 
  "extract_proof (Inr (TRS_Proof _ _ _ prf)) = Some prf"
| "extract_proof _ = None"

fun extract_trs where
  "extract_trs (Inr (TRS_Proof _ trs _ _)) = Some trs"
| "extract_trs _ = None"

definition my_trs :: "((string,label_type)lab,string)rules" where "my_trs \<equiv> the (extract_trs my_proof)"

definition div_trs where "div_trs \<equiv>  [
     (Fun ''id'' [Var ''x''], Var ''x''), 
     (Fun ''minus'' [Var ''x'', Fun ''0'' []], Var ''x''),
     (Fun ''minus'' [Fun ''s'' [Var ''x''], Fun ''s'' [Var ''y'']], Fun ''minus'' [Var ''x'', Var ''y'']),
     (Fun ''div'' [Fun ''0'' [], Fun ''s'' [Var ''y'']], Fun ''0'' []),
     (Fun ''div'' [Fun ''s'' [Var ''x''], Fun ''s'' [Var ''y'']],
          Fun ''s'' [Fun ''div'' [Fun ''minus'' [Var ''x'', Var ''y''], Fun ''id'' [Fun ''s'' [Var ''y'']]]])]"

lemma sn_div_trs: "SN (rstep (set div_trs))"
proof (rule sn_via_ceta[of _ "the (extract_proof my_proof)"])
  show "sn_via_ceta div_trs (the (extract_proof my_proof))"
    apply (simp add: div_trs_def my_proof_def)
    apply (unfold dependency_pairs_tt_def Let_def tp_rbt_impl_def)
    apply (simp add: check_wf_trs_def check_def check_return_def)
    apply (simp add: Termination_Problem_Impl.mk_impl_def)
    apply (simp add: check_subseteq_def vars_term_list.simps vars_term_list'.simps)
    apply (simp add: DP_list_def map_trs_list.simps Let_def supt_impl supteq_impl)
    apply (simp add: mem_def)
    apply (simp add:  computeRuleMap_def reverse_trs_def del: lookup.simps)
    apply (simp only: checkDPProof.simps debug.simps)
    apply (simp add: check_dep_graph_proc.simps decomp.simps check_allm_gen_index_def check_edges.simps capRM.simps check_def gcMatchI.simps check_subseteq_def del: gcMatchISound)    
    apply (unfold checkDPProof.simps(3)[of "((''1'' +#+ shows_string ''.1'') +@+ ''.'' +#+ shows (Suc 0))"])
    apply (simp only: debug.simps)
    apply (simp only: check_redpair_ur_af_proc.simps)
    apply (simp)
    apply (simp add: create_poly_redpair_def)
    apply (simp add: check_def)
    apply (simp add: Let_def)
    apply (simp add: check_lpoly_coeffs.simps)
    apply (simp add: check_ur_P_closed_rm_af.simps check_def)
    apply (simp add: create_afI.simps)
    apply (simp add: forallm_index_def check_ur_closed_term_rm_af.simps)
    apply (simp add: check_def check_subseteq_def)
    apply (unfold checkDPProof.simps(1))
    apply (simp add: debug.simps)
    apply (unfold checkDPProof.simps(6))
    apply (simp add: debug.simps del: check_mono_redpair_ur_proc.simps)
    apply (unfold check_mono_redpair_ur_proc.simps)
    apply (simp add: create_poly_redpair_def)
    apply (simp add: check_def)
    apply (simp add: Let_def)
    apply (simp add: check_lpoly_coeffs.simps)
    apply (simp add: check_ur_P_closed_rm_af.simps check_def)
    apply (simp add: create_afI.simps int_mono_def)
    apply (simp add: forallm_index_def \<pi>_full_def check_ur_closed_term_rm_af.simps)
    apply (simp add: check_def check_subseteq_def)
    apply (simp add: checkDPProof.simps)
    apply (simp add: create_poly_redpair_def check_def)
    apply (simp add: Let_def int_mono_def)
    apply (simp add: check_lpoly_coeffs.simps)
    apply (simp add: check_def)
    apply (simp add: check_subseteq_def)
  done
qed



ML {*
structure CetaRules = Named_Thms
  (val name = "ceta"
   val description = "Theorems for execution of CeTA")
*}

setup {* CetaRules.setup *}

ML {*
  Code_Preproc.print_codeproc @{theory}
*}

(* /Applications/Isabelle2009-2.app/Isabelle/src/HOL/Code_Evaluation.thy *)
(* /Applications/Isabelle2009-2.app/Isabelle/src/Tools/Code/code_preproc.ML *)

lemmas ceta_congs = sum.weak_case_cong if_weak_cong list.weak_case_cong nat.weak_case_cong

lemma ceta_list: shows
   [ceta]: "length [] = 0" and [ceta]: "\<And> x xs. length (x # xs) = Suc (length xs)" 
  and [ceta]: "null []" and [ceta]: "\<And> x xy. \<not> (null (x # xs))" 
  and [ceta]: "\<And> x xy. (x # xs) \<noteq> []" by auto

lemmas xml_parser_simps = wspace_def letters_def dquote_def splitter_fun.simps splitter_def trim.simps bind_def 
  update_tokens_def get_tokens_def set_tokens_def quote_def expecting_def eoi.simps oneof_aux.simps oneof_def many.simps
  manyof_def spaces_def scan_upto.simps parse_name_def parse_text_def parse_attribute_value_def
  parse_attributes.simps parse_nodes.simps parse_node_def parse_header.simps remove_comments_aux.simps remove_comments_def
  parse_doc_def doc_of_string_def debug.simps nat_number take.simps
  sum.cases list.cases nat.cases Nat.nat.simps  add_Suc member_simps
  eq_nat_number_of neg_number_of_Bit0 neg_number_of_Bit1 numeral_1_eq_1 not_neg_1 numeral_0_eq_0 not_neg_0 add.left_neutral
  One_nat_def
  if_True if_False map.simps zip.simps Error_Monad.bind_def
  Let_def list.inject char.inject nibble.simps HOL.simp_thms newline_def drop.simps shows_between_def newline_def
  shows_prec_list_def shows_list_list_def shows_prec_char_def shows_list_char_def split_def fst_conv snd_conv id_def
  List.append.simps shows_string_def shows_list_aux_def shows_list_gen_def shows_sep.simps

declare xml_parser_simps[ceta]

lemma "(if (id x) then id 1 else id 2) = id 3"
proof (tactic {* simp_tac (HOL_basic_ss addsimps @{thms ceta} addcongs @{thms ceta_congs}) 1 *})
oops


(* ceta-simps now suffice to simplify doc_of_string *)

definition test_term where  "test_term == doc_of_string ''<?xml version=\"1.0\"?><certificationProblem><foo/></certificationProblem>''" 
(* definition test_term where  "test_term == doc_of_string ''<ba><fo/></ba>''"*)
term parseCertProblem
term doc_of_string

lemma "test_term = x" proof (simp only: ceta test_term_def cong: ceta_congs)
oops
ML {*
  op ::
*}
ML {*
  val t = Eval.eval_term @{theory} @{term test_term} |> cterm_of @{theory}; 
  val my_simpset = HOL_basic_ss addsimps (@{thm test_term_def} :: @{thms ceta}) addcongs @{thms ceta_congs};
  val s = timeit (fn () => Simplifier.rewrite my_simpset @{cterm test_term})
*}
ML {*
  fun inp x = (* "some_prover " ^ x
            |> bash_output
            |> fst *)
            File.read (Path.explode x)
            |> HOLogic.mk_string
            |> (fn t => @{term certifyProof} $ t)    
(*            |> (fn t => @{term doc_of_string} $ t)   *)
(*            |> (fn t => @{term parseCertProblem} $ t)   *)
            |> Eval.eval_term @{theory}; 
(*            |> cterm_of @{theory} |> Simplifier.rewrite my_simpset |> prop_of |> (fn (_ $ y) => y);  *)
    fun import_term file  =  snd o Local_Theory.define ((@{binding my_term}, NoSyn), ((@{binding my_term_def}, []), inp file) )
*}

(*thm if_weak_cong
lemma "(if x then id 1 else id 2) = id 3" proof (simp split del: split_if)
*)

local_setup {*
  timeit (fn _ => import_term "~/rewriting/examples/defined_arity.proof.xml" (* "test.xml" *) )
*}
term my_term
thm my_term_def

end

