(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Term_Multiset
imports
  Term
  Term_Impl
  Multiset2
begin

text {*
  here we give some function to compute multisets instead of sets of 
  function symbols, variables, etc.
  they may be helpful for non-duplicating TRSs 
*}

fun funs_term_ms :: "('f,'v)term \<Rightarrow> 'f multiset"
  where "funs_term_ms (Var x) = {#}"
     |  "funs_term_ms (Fun f ts) = {#f#} + Multiset2.concat (multiset_of (map funs_term_ms ts))"

fun funs_ctxt_ms :: "('f,'v)ctxt \<Rightarrow> 'f multiset"
  where "funs_ctxt_ms Hole = {#}"
     |  "funs_ctxt_ms (More f bef C aft) = {#f#} + Multiset2.concat (multiset_of (map funs_term_ms bef)) + funs_ctxt_ms C + Multiset2.concat (multiset_of (map funs_term_ms aft))"

lemma funs_ms_ctxt_apply: "funs_term_ms C\<langle>t\<rangle> = funs_ctxt_ms C + funs_term_ms t"
  by (induct C, auto simp: multiset_eq_iff)

fun vars_term_ms :: "('f,'v)term \<Rightarrow> 'v multiset"
  where "vars_term_ms (Var x) = {#x#}"
     |  "vars_term_ms (Fun f ts) = Multiset2.concat (multiset_of (map vars_term_ms ts))"

lemma funs_ms_subst_apply: 
  "funs_term_ms (t \<cdot> \<sigma>) = funs_term_ms t + 
            Multiset2.concat (Multiset2.map (\<lambda> x. funs_term_ms (\<sigma> x)) (vars_term_ms t))" 
proof (induct t)
  case (Var x) show ?case by auto
next
  case (Fun f ts)
  let ?m = "multiset_of"
  let ?f = "funs_term_ms"
  let ?v = "vars_term_ms"
  let ?ts = "Multiset2.concat (?m (map ?f ts))"
  let ?\<sigma>  = "Multiset2.concat (Multiset2.map (\<lambda>x. ?f (\<sigma> x)) (Multiset2.concat (?m (map ?v ts))))"
  let ?ts\<sigma> = "Multiset2.concat (?m (map (\<lambda> x. ?f (x \<cdot> \<sigma>)) ts))"
  have ind: "?ts\<sigma> = ?ts + ?\<sigma>" using Fun 
  proof (induct ts)
    case Nil show ?case by simp
  next
    case (Cons t ts)
    show ?case 
      by (simp add: Cons, unfold multiset_eq_iff, auto)
  qed
  thus ?case
    by (simp add: o_def, unfold multiset_eq_iff, auto)
qed



lemma multiset_of_funs_term_list': "multiset_of (funs_term_list' t) = funs_term_ms t"
proof (induct t)
  case (Var x) show ?case by (simp add: funs_term_list'.simps)
next
  case (Fun f ts)
  show ?case 
    by (simp add: funs_term_list'.simps, insert Fun, induct ts, auto simp: funs_term_list'.simps multiset_eq_iff)
qed

lemma multiset_of_vars_term_list': "multiset_of (vars_term_list' t) = vars_term_ms t"
proof (induct t)
  case (Var x) show ?case by (simp add: vars_term_list'.simps)
next
  case (Fun f ts)
  show ?case 
    by (simp add: vars_term_list'.simps, insert Fun, induct ts, auto simp: vars_term_list'.simps multiset_eq_iff)
qed

lemma set_of_vars_term_ms[simp]: "set_of (vars_term_ms t) = vars_term t"
  unfolding multiset_of_vars_term_list'[symmetric] vars_term_list'_sound[symmetric]
  by auto

lemma ground_vars_term_ms_empty: "ground t = (vars_term_ms t = {#})"
  unfolding ground_vars_term_empty
  unfolding set_of_vars_term_ms[symmetric]
  by (simp del: set_of_vars_term_ms)



lemma vars_term_ms_map_term[simp]: "vars_term_ms (map_term fg t) = vars_term_ms t"
proof (induct t)
  case (Var x) thus ?case by simp
next
  case (Fun f ts) 
  show ?case
    by (simp, insert Fun, induct ts, auto)
qed

lemma linear_vars_term_ms: assumes "linear_term t"
  shows "count (vars_term_ms t) x <= 1"
  unfolding multiset_of_vars_term_list'[symmetric]
  unfolding count_multiset_of
  using linear_vars_term_list'[OF assms] by auto

lemma right_linear_imp_non_duplicating: assumes rl: "linear_term r"
  and vc: "vars_term r \<subseteq> vars_term l"
  shows "vars_term_ms r <= vars_term_ms l"
proof -
  let ?L = "vars_term_ms l"
  let ?R = "vars_term_ms r"
  let ?M = "?L - ?R"
  let ?N = "set_of ?L"
  obtain M where M: "M = ?M" by auto
  show ?thesis 
  proof (rule mset_less_eqI)
    fix x
    show "count ?R x <= count ?L x"
    proof (cases "count ?R x")
      case 0
      thus ?thesis by auto
    next
      case (Suc n)
      with linear_vars_term_ms[OF rl, of x] have one: "count (vars_term_ms r) x = 1"
        by auto
      hence "x \<in># ?R" by auto
      hence "x \<in> set_of ?R" unfolding mem_set_of_iff .
      hence "x \<in> vars_term r" by simp
      with vc have "x \<in> vars_term l" by auto
      hence "x \<in> set_of ?L" by auto
      hence "x \<in># ?L" unfolding mem_set_of_iff .
      hence "0 < count ?L x" by auto
      with one show ?thesis by auto
    qed
  qed
qed

lemma vars_term_ms_subst[simp]: 
  "vars_term_ms (t \<cdot> \<sigma>) = Multiset2.concat (Multiset2.map (\<lambda> x. vars_term_ms (\<sigma> x)) (vars_term_ms t))" (is "_ = ?V t")
proof (induct t)
  case (Fun f ts)
  have IH: "map (\<lambda> t. vars_term_ms (t \<cdot> \<sigma>)) ts = map (\<lambda> t. ?V t) ts"
    by (rule map_cong[OF refl Fun])
  show ?case by (simp add: o_def IH, induct ts, auto)
qed simp


lemma vars_term_ms_subst_mono: assumes "vars_term_ms s \<le> vars_term_ms t" 
  shows "vars_term_ms (s \<cdot> \<sigma>) \<le> vars_term_ms (t \<cdot> \<sigma>)"
proof -
  from assms[unfolded mset_le_exists_conv] obtain u where t: "vars_term_ms t = vars_term_ms s + u" by auto
  show ?thesis unfolding vars_term_ms_subst unfolding t by auto
qed


lemma supteq_imp_vars_term_ms_subset: assumes "s \<unrhd> t"
  shows "vars_term_ms t \<le> vars_term_ms s" 
proof -
  from assms have "\<exists> bef aft. vars_term_list' s = bef @ vars_term_list' t @ aft"
  proof (induct rule: supteq.induct)
    case (refl t)
    show ?case by ((rule exI[of _ Nil])+, auto)
  next
    case (subt u ss t f)
    from subt(3) obtain bef aft where u: "vars_term_list' u = bef @ vars_term_list' t @ aft" by auto
    from subt(1)[unfolded in_set_conv_decomp] obtain bbef aaft where ss: "ss = bbef @ u # aaft" by auto
    show ?case
      by (rule exI[of _ "concat (map vars_term_list' bbef) @ bef"], unfold ss,
      auto simp: vars_term_list'.simps u)
  qed
  then obtain bef aft where id: "vars_term_list' s = bef @ vars_term_list' t @ aft" by blast
  show ?thesis unfolding multiset_of_vars_term_list'[symmetric] id
    using mset_le_add_right[of "multiset_of (vars_term_list' t)" "multiset_of (bef @ aft)"] by (auto simp: ac_simps)
qed

end
